@@ -328,7 +328,8 @@ State *DFA::add_new_state(optflags const &opts, NodeSet *anodes,
 	proto.init(nnodev, anodev);
 	State *state;
 	try {
-		state = new State(opts, node_map.size(), proto, other, filedfa);
+		state = new State(uniq_perms, opts, node_map.size(), proto,
+				  other, filedfa);
 	} catch(int error) {
 		/* this function is called in the DFA object creation,
 		 * and the exception prevents the destructor from
@@ -559,18 +560,13 @@ State *DFA::match(const char *str)
 
 void DFA::dump_uniq_perms(const char *s)
 {
-	set<perms_t> uniq;
-	for (Partition::iterator i = states.begin(); i != states.end(); i++)
-		uniq.insert((*i)->perms);
-
-	cerr << "Unique Permission sets: " << s << " (" << uniq.size() << ")\n";
+	cerr << "Unique Permission sets: " << s << " (" << uniq_perms.size() << ")\n";
 	cerr << "----------------------\n";
-	for (set<perms_t >::iterator i = uniq.begin(); i != uniq.end(); i++) {
-		cerr << "  allow:" << hex << i->allow << " deny:"
-		     << i->deny << " audit:" << i->audit
-		     << " quiet:" << i->quiet << dec << "\n";
+	for (std::set<perms_t *>::iterator i = uniq_perms.begin(); i != uniq_perms.end(); i++) {
+	  cerr << "  allow:" << hex << (*i)->allow << " deny:"
+	       << (*i)->deny << " audit:" << (*i)->audit
+	       << " quiet:" << (*i)->quiet << " prompt:" << (*i)->prompt <<  dec << "\n";
 	}
-	//TODO: add prompt
 }
 
 /* Remove dead or unreachable states */
@@ -609,8 +605,8 @@ void DFA::remove_unreachable(optflags const &opts)
 					cerr << "unreachable: " << **i;
 					if (*i == start)
 						cerr << " <==";
-					if ((*i)->perms.is_accept())
-						(*i)->perms.dump(cerr);
+					if ((*i)->perms->is_accept())
+						(*i)->perms->dump(cerr);
 					cerr << "\n";
 				}
 				State *current = *i;
@@ -672,17 +668,20 @@ bool DFA::same_mappings(State *s1, State *s2)
 int DFA::apply_and_clear_deny(void)
 {
 	int c = 0;
+	/* TODO: update to remove perms that are no longer in use */
 	for (Partition::iterator i = states.begin(); i != states.end(); i++)
-		c += (*i)->apply_and_clear_deny();
+		c += (*i)->apply_and_clear_deny(uniq_perms);
 
 	return c;
 }
 
 
+typedef map<perms_t *, Partition *, deref_less_than_perms> PermMap;
+
 /* minimize the number of dfa states */
 void DFA::minimize(optflags const &opts)
 {
-	map<perms_t, Partition *> perm_map;
+	PermMap perm_map;
 	list<Partition *> partitions;
 
 	/* Set up the initial partitions
@@ -691,14 +690,14 @@ void DFA::minimize(optflags const &opts)
 	int accept_count = 0;
 	int final_accept = 0;
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		map<perms_t, Partition *>::iterator p = perm_map.find((*i)->perms);
+		PermMap::iterator p = perm_map.find((*i)->perms);
 		if (p == perm_map.end()) {
 			Partition *part = new Partition();
 			part->push_back(*i);
 			perm_map.insert(make_pair((*i)->perms, part));
 			partitions.push_back(part);
 			(*i)->partition = part;
-			if ((*i)->perms.is_accept())
+			if ((*i)->perms->is_accept())
 				accept_count++;
 		} else {
 			(*i)->partition = p->second;
@@ -810,9 +809,23 @@ void DFA::minimize(optflags const &opts)
 			if (opts.dump & DUMP_DFA_MIN_PARTS)
 				cerr << **i << ", ";
 			(*i)->label = -1;
-			rep->perms.add((*i)->perms, filedfa);
+			/* merging perms is only necessary if partitioning doesn't
+			 * completely separate base on unique perms.
+			 * atm this should be the case. Code to handle case is
+			 * only to document what should be done if this is allowed
+			 * in the future
+			 */
+			if ((rep->perms != (*i)->perms) &&
+			    (*rep->perms != *(*i)->perms)) {
+				throw std::runtime_error("Minimization different permissions in same partion");
+				/*
+				perms_t tmp = *rep->perms;
+				tmp.add((*i)->perms, filedfa);
+				rep->perms = uniq_perms.insert(tmp);
+				*/
+			}
 		}
-		if (rep->perms.is_accept())
+		if (rep->perms->is_accept())
 			final_accept++;
 		if (opts.dump & DUMP_DFA_MIN_PARTS)
 			cerr << "\n";
@@ -1098,14 +1111,14 @@ void DFA::dump_diff_encode(ostream &os)
 void DFA::dump(ostream &os, Renumber_Map *renum)
 {
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
-		if (*i == start || (*i)->perms.is_accept()) {
+		if (*i == start || (*i)->perms->is_accept()) {
 			os << make_pair(*i, renum);
 			if (*i == start) {
 				os << " <== ";
-				(*i)->perms.dump_header(os);
+				(*i)->perms->dump_header(os);
 			}
-			if ((*i)->perms.is_accept())
-				(*i)->perms.dump(os);
+			if ((*i)->perms->is_accept())
+				(*i)->perms->dump(os);
 			os << "\n";
 		}
 	}
@@ -1123,16 +1136,16 @@ void DFA::dump(ostream &os, Renumber_Map *renum)
 				if (first) {
 					first = false;
 					os << make_pair(*i, renum) << " perms: ";
-					if ((*i)->perms.is_accept())
-						(*i)->perms.dump(os);
+					if ((*i)->perms->is_accept())
+						(*i)->perms->dump(os);
 					else
 						os << "none";
 					os << "\n";
 				}
 				os << "    "; j->first.dump(os) << " -> " <<
 					make_pair(j->second, renum);
-				if ((j)->second->perms.is_accept())
-					os << " ", (j->second)->perms.dump(os);
+				if ((j)->second->perms->is_accept())
+					os << " ", (j->second)->perms->dump(os);
 				os << "\n";
 			}
 		}
@@ -1141,8 +1154,8 @@ void DFA::dump(ostream &os, Renumber_Map *renum)
 			if (first) {
 				first = false;
 				os << make_pair(*i, renum) << " perms: ";
-				if ((*i)->perms.is_accept())
-					(*i)->perms.dump(os);
+				if ((*i)->perms->is_accept())
+					(*i)->perms->dump(os);
 				else
 					os << "none";
 				os << "\n";
@@ -1156,8 +1169,8 @@ void DFA::dump(ostream &os, Renumber_Map *renum)
 				}
 			}
 			os << "] -> " << make_pair((*i)->otherwise, renum);
-			if ((*i)->otherwise->perms.is_accept())
-				os << " ", (*i)->otherwise->perms.dump(os);
+			if ((*i)->otherwise->perms->is_accept())
+				os << " ", (*i)->otherwise->perms->dump(os);
 			os << "\n";
 		}
 	}
@@ -1179,9 +1192,9 @@ void DFA::dump_dot_graph(ostream & os)
 		if (*i == start) {
 			os << "\t\tstyle=bold" << "\n";
 		}
-		if ((*i)->perms.is_accept()) {
+		if ((*i)->perms->is_accept()) {
 			os << "\t\tlabel=\"" << **i << "\\n";
-			(*i)->perms.dump(os);
+			(*i)->perms->dump(os);
 			os << "\"\n";
 		}
 		os << "\t]" << "\n";
@@ -1327,41 +1340,58 @@ void DFA::apply_equivalence_classes(map<transchar, transchar> &eq)
 	}
 }
 
-void DFA::compute_perms_table_ent(State *state, size_t pos,
-				  vector <aa_perms> &perms_table,
-				  bool prompt)
+void DFA::compute_perms_table_ent(perms_t * const perms, size_t pos,
+				  std::vector <aa_perms> &perms_table,
+				  idxmap_t &idxmap, bool prompt)
 {
 	uint32_t accept1, accept2, accept3;
 
 	// until front end doesn't map the way it does
-	state->map_perms_to_accept(accept1, accept2, accept3, prompt);
+	  perms->map_perms_to_accept(accept1, accept2, accept3, prompt);
+	idxmap.insert(make_pair(perms, pos));
 	if (filedfa) {
-		state->idx = pos * 2;
-		perms_table[pos*2] = compute_fperms_user(accept1, accept2, accept3);
-		perms_table[pos*2 + 1] = compute_fperms_other(accept1, accept2, accept3);
+		perms_table[pos] = compute_fperms_user(accept1, accept2, accept3);
+		perms_table[pos + 1] = compute_fperms_other(accept1, accept2, accept3);
 	} else {
-		state->idx = pos;
 		perms_table[pos] = compute_perms_entry(accept1, accept2, accept3);
 	}
 }
 
 void DFA::compute_perms_table(vector <aa_perms> &perms_table, bool prompt)
 {
+	idxmap_t idxmap;
+
 	size_t mult = filedfa ? 2 : 1;
-	size_t pos = 2;
+	size_t pos = filedfa ? 2 : 1;
 
 	assert(states.size() >= 2);
-	perms_table.resize(states.size() * mult);
+
+	perms_table.resize(uniq_perms.size()*mult);
+	compute_perms_table_ent(nonmatching->perms, 0, perms_table, idxmap,
+				prompt);
+	nonmatching->idx = 0;
+	start->idx = 0;
+
+	for (perms_t_Cache::const_iterator i = uniq_perms.cbegin(); i != uniq_perms.cend(); i++) {
+		if (*i == nonmatching->perms)
+			continue;
+		compute_perms_table_ent(*i, pos, perms_table, idxmap, prompt);
+		pos += mult;
+	}
 
 	// nonmatching and start need to be 0 and 1 so handle outside of loop
-	compute_perms_table_ent(nonmatching, 0, perms_table, prompt);
-	compute_perms_table_ent(start, 1, perms_table, prompt);
-
 	for (Partition::iterator i = states.begin(); i != states.end(); i++) {
 		if (*i == nonmatching || *i == start)
 			continue;
-		compute_perms_table_ent(*i, pos, perms_table, prompt);
-		pos++;
+		idxmap_t::iterator j = idxmap.find((*i)->perms);
+		if (j == idxmap.end()) {
+			perms_t_Cache::iterator k = uniq_perms.find((*i)->perms);
+			if (k == uniq_perms.end())
+				throw std::runtime_error("permission not in permission table map");
+			else
+				throw std::runtime_error("permission not in idx table map");
+		}
+		(*i)->idx = j->second;
 	}
 }
 
@@ -1555,16 +1585,15 @@ static int pri_update_perm(optflags const &opts, vector<int> &priority, int i,
  * have any exact matches, then they override the execute and safe
  * execute flags.
  */
-int accept_perms(optflags const &opts, NodeVec *state, perms_t &perms,
-		 bool filedfa)
+perms_t::perms_t(optflags const &opts, NodeVec *state, bool filedfa)
 {
 	int error = 0;
 	// scaling priority by *4
-	std::vector<int>  priority(sizeof(perm32_t)*8,  MIN_INTERNAL_PRIORITY<<2);	// 32 but wan't tied to perm32_t
-	perms.clear();
+	std::vector<int>  priority(sizeof(perm32_t)*8,  MIN_INTERNAL_PRIORITY*4);	// 32 but wasn't tied to perm32_t
+	clear();
 
 	if (!state)
-		return error;
+		return;
 	if (opts.dump & DUMP_DFA_PERMS) {
 		cerr << "Building Perms";
 		if (filedfa)
@@ -1584,7 +1613,7 @@ int accept_perms(optflags const &opts, NodeVec *state, perms_t &perms,
 
 		for (int i = 0;  check; i++) {
 			if (check & bit) {
-				error = pri_update_perm(opts, priority, i, match, perms, filedfa);
+				error = pri_update_perm(opts, priority, i, match, *this, filedfa);
 				if (error)
 					goto out;
 			}
@@ -1592,15 +1621,15 @@ int accept_perms(optflags const &opts, NodeVec *state, perms_t &perms,
 			bit <<= 1;
 		}
 	}
-	if (filedfa && (perms.allow & AA_EXEC_BITS)) {
-		add_implied_ix_mmap(opts, priority, perms.allow);
+	if (filedfa && (allow & AA_EXEC_BITS)) {
+		add_implied_ix_mmap(opts, priority, allow);
 	}
 	if (opts.dump & DUMP_DFA_PERMS) {
-		cerr << "  computed: ";  perms.dump(cerr); cerr << "\n";
+		cerr << "  computed: ";  dump(cerr); cerr << "\n";
 	}
 out:
-	if (error)
+	if (error) {
 		fprintf(stderr, "profile has merged rule with conflicting x modifiers\n");
-
-	return error;
+		throw error;
+	}
 }
